import sys

class Logging:

    #---------------------------------------------------------------------------

    levels = {
        # this isn't a real message level, it's just for purposes of being
        # able to turn logging off entirely
        'OFF': {
            'severity': 10
        },
        'ERROR': {
            'severity': 5,
            'prefix': '! ',
            'fatal': True
        },
        'WARN': {
            'severity': 4,
            'prefix': '! ',
        },
        'INFO': {
            'severity': 3,
            'prefix': '',
        },
        'DEBUG': {
            'severity': 2,
            'prefix': '  %% ',
        },
        'TRACE': {
            'severity': 1,
            'prefix': '    ++ ',
        }
    }

    # This is a fall-back default; it should only be relevant if an invalid
    # logging level is specified on the command line.
    default_logging_level = 'INFO'

    #---------------------------------------------------------------------------

    @classmethod
    def set_logging_level(cls, level):
        if level in cls.levels:
            cls.logging_level = level
        else:
            cls.message('WARN', 'Attempt to set logging level to unknown level ' + level + ', using default value ' + cls.default_logging_level + ' instead')
            cls.logging_level = cls.default_logging_level
    
    #---------------------------------------------------------------------------

    @classmethod
    def print_to_console (cls, level, text):
        print(cls.levels[level]['prefix'] + text, file = sys.stderr)
    
    #---------------------------------------------------------------------------
    
    @classmethod
    def message (cls, level, text):
        if level not in cls.levels:
            cls.print_to_console('WARN', 'Invalid log message level ' + level)
        else:
            if cls.levels[level]['severity'] >= cls.levels[cls.logging_level]['severity']:
                cls.print_to_console(level, text)
            # note that if a level is set as fatal, we terminate the program
            # regardless of the logging level
            if 'fatal' in cls.levels[level] and cls.levels[level]['fatal']:
                sys.exit(1)

    #---------------------------------------------------------------------------

logging = Logging



from datetime import datetime
import os.path
import json
from salal.core.logging import logging
from salal.core.config import config
from salal.core.utilities import utilities

class DependencyManager:

    separator = ':'
    
    #---------------------------------------------------------------------------

    @classmethod
    def initialize (cls):
        logging.message('DEBUG', 'Initializing dependency tracking')

        # read the build log
        cls.build_log_file = os.path.join(config.system['paths']['config_root'], config.system['paths']['build_log_dir'], config.system['profile'] + '.json')
        if os.path.isfile(cls.build_log_file):
            with open(cls.build_log_file, 'r') as build_log_fh:
                build_log = json.load(build_log_fh)
            cls.last_build_time = datetime.strptime(build_log['timestamp'], '%Y-%m-%d %H:%M:%S.%f')
            cls.file_log = build_log['files']
            cls.variable_log = build_log['variables']
            cls.template_log = set(build_log['templates'])
            cls.resource_log = set(build_log['resources'])
        else:
            cls.last_build_time = datetime.min
            cls.file_log = dict()
            cls.variable_log = dict()
            cls.template_log = set()
            cls.resource_log = set()

        # initialize utility variables
        cls.n_files_checked = 0
        cls.n_files_built = 0
        cls.cur_target = None
        cls.cur_source = None
        cls.cur_file_key = None
            
        # initialize the update queue
        cls.cur_build_time = datetime.now()
        cls.file_updates = dict()
        cls.variable_updates = dict()
        cls.template_updates = set()
        cls.resource_updates = set()

        cls.file_check_flags = {file_ref:False for file_ref in cls.file_log.keys()}
        cls.variable_change_flags = cls.check_for_variable_changes(cls.variable_log, cls.variable_updates)
        cls.template_change_flags = cls.check_for_file_changes(cls.template_log, 'template')
        cls.resource_change_flags = cls.check_for_file_changes(cls.resource_log, 'resource')
    
    #---------------------------------------------------------------------------
    # For each variable in the log, we create a flag to indicate
    # whether the variabe has changed since the last build, which is
    # used to help determine when a source file should be rebuilt.
    @classmethod
    def check_for_variable_changes (cls, variable_log, variable_updates):
        variable_change_flags = dict()
        for variable in variable_log:
            # It's not necessarily an error if a variable from the
            # last build doesn't exist any more, because references to
            # that variable may also have been removed. So we don't
            # throw an error here, but we do warn and trigger a
            # rebuild of any source files that referenced the
            # variable.
            if not variable in config.project:
                logging.message('WARN', 'Variable ' + variable + ' is in the build log but no longer exists')
                variable_change_flags[variable] = True
            elif variable_log[variable] != config.project[variable]:
                logging.message('TRACE', 'Detected change to variable ' + variable)
                variable_change_flags[variable] = True
                variable_updates[variable] = config.project[variable]
            else:
                variable_change_flags[variable] = False
        return variable_change_flags
                    
    #---------------------------------------------------------------------------

    # For each file in the <file_list>, we create a flag to indicate
    # whether the file has changed since the last build, which is
    # used to help determine when a source file should be rebuilt.
    @classmethod
    def check_for_file_changes (cls, file_list, file_type):
        file_change_flags = dict()
        for file_path in file_list:
            # It's not necessarily an error if a file from the
            # last build doesn't exist any more, because references to
            # that file may also have been removed. So we don't
            # throw an error here, but we do warn trigger a rebuild of
            # any source files that referenced the file.
            if not os.path.isfile(file_path):
                logging.message('WARN', file_type.capitalize() + ' ' + file_path + ' is in the build log but no longer exists')
                file_change_flags[file_path] = True
            else:
                file_mod_time = datetime.fromtimestamp(os.path.getmtime(file_path))
                if file_mod_time > cls.last_build_time:
                    logging.message('TRACE', 'Detected change to ' + file_type + ' ' + file_path)
                    file_change_flags[file_path] = True
                else:
                    file_change_flags[file_path] = False
        return file_change_flags
                    
    #---------------------------------------------------------------------------
    
    @classmethod
    def needs_build (cls, target_file, source_file):
        cls.n_files_checked += 1
        target_key = target_file + cls.separator + source_file
        
        # Is this file in the build log? If not, rebuild, if yes, then
        # record that a build check was conducted for it and proceed.
        if target_key not in cls.file_log:
            logging.message('TRACE', 'Target ' + target_file + ' is not in the build log, build required')
            return True
        else:
            cls.file_check_flags[target_key] = True

        # Does the target exist? If not, rebuild.
        if not os.path.isfile(target_file):
            logging.message('TRACE', 'Target ' + target_file + ' does not exist, build required')
            return True

        # Is the source newer than the target? If so, rebuild.
        source_mod_time = datetime.fromtimestamp(os.path.getmtime(source_file))
        if source_mod_time > cls.last_build_time:
            logging.message('TRACE', 'Source file ' + source_file + ' is newer than target ' + target_file + ', build required')
            return True

        # Have any referenced variables been changed? If so, rebuild.
        for variable in cls.file_log[target_key]['variables']:
            if cls.variable_change_flags[variable]:
                logging.message('TRACE', 'Variable ' + variable + ' used by target ' + target_file + ' has changed, build required')
                return True

        # Have any referenced templates been changed? If so, rebuild.
        for template in cls.file_log[target_key]['templates']:
            if cls.template_change_flags[template]:
                logging.message('TRACE', 'Template ' + template + ' used by target ' + target_file + ' has changed, build required')
                return True

        # Have any referenced resources been changed? If so, rebuild.
        for resource in cls.file_log[target_key]['resources']:
            if cls.resource_change_flags[resource]:
                logging.message('TRACE', 'Resource ' + resource + ' used by target ' + target_file + ' has changed, build required')
                return True

        return False

    #---------------------------------------------------------------------------

    @classmethod
    def start_build_tracking (cls, target_file, source_file):
        cls.n_files_built += 1
        cls.cur_target = target_file
        cls.cur_source = source_file
        cls.cur_file_key = target_file + cls.separator + source_file
        if cls.cur_file_key not in cls.file_log:
            logging.message('TRACE', 'Detected new build target ' + target_file + ', now tracking it')
        cls.file_updates[cls.cur_file_key] = {
            'target': target_file,
            'source': source_file,
            'variables': [],
            'templates': [],
            'resources': []
        }
    
    #---------------------------------------------------------------------------

    @classmethod
    def variable_used (cls, variable_name):
        if variable_name not in cls.file_updates[cls.cur_file_key]['variables']:
            cls.file_updates[cls.cur_file_key]['variables'].append(variable_name)
            if variable_name not in cls.variable_log and variable_name not in cls.variable_updates:
                logging.message('TRACE', 'Detected use of new variable ' + variable_name + ', now tracking it');
                cls.variable_updates[variable_name] = config.project[variable_name] 

    #---------------------------------------------------------------------------

    @classmethod
    def variable_not_found (cls, variable_name):
        logging.message('WARN', 'Encountered reference to undefined variable ' + variable_name)
    
    #---------------------------------------------------------------------------
    
    @classmethod
    def template_used (cls, template_file):
        if template_file not in cls.file_updates[cls.cur_file_key]['templates']:
            cls.file_updates[cls.cur_file_key]['templates'].append(template_file)
            if template_file not in cls.template_log and template_file not in cls.template_updates:
                logging.message('TRACE', 'Detected use of new template ' + template_file + ', now tracking it');
                cls.template_updates.add(template_file) 
                
    #---------------------------------------------------------------------------
    
    @classmethod
    def resource_used (cls, resource_file):
        if resource_file not in cls.file_updates[cls.cur_file_key]['resources']:
            cls.file_updates[cls.cur_file_key]['resources'].append(resource_file)
            if resource_file not in cls.resource_log and resource_file not in cls.resource_updates:
                logging.message('TRACE', 'Detected use of new resource ' + resource_file + ', now tracking it');
                cls.resource_updates.add(resource_file) 
                
    #---------------------------------------------------------------------------

    @classmethod
    def stop_build_tracking (cls):
        cls.cur_target = None
        cls.cur_source = None
        cls.cur_file_key = None
        
    #---------------------------------------------------------------------------

    @classmethod
    def num_files_checked (cls):
        return cls.n_files_checked
    
    #---------------------------------------------------------------------------
    
    @classmethod
    def num_files_built (cls):
        return cls.n_files_built
    
    #---------------------------------------------------------------------------
    
    @classmethod
    def remove_stale_references (cls, log, change_flags, reference_type):
        active_references = set()
        for file_entry in cls.file_log.values():
            if reference_type + 's' in file_entry:
                for reference_path in file_entry[reference_type + 's']:
                    active_references.add(reference_path)
        for reference_path in change_flags:
            if reference_path not in active_references:
                logging.message('TRACE', reference_type.capitalize() + ' ' + reference_path + ' is no longer part of the build, discontinuing tracking');
                log.remove(reference_path)
                
    #---------------------------------------------------------------------------
    
    @classmethod
    def write_log (cls):
        logging.message('DEBUG', 'Updating build log')
        
        # incorporate updates to the file, variable, template, and resource logs
        cls.file_log.update(cls.file_updates)
        cls.variable_log.update(cls.variable_updates)
        cls.template_log.update(cls.template_updates)
        cls.resource_log.update(cls.resource_updates)

        # remove entries for stale files (were in build log but no
        # longer part of the build)
        for file_ref in cls.file_check_flags:
            if not cls.file_check_flags[file_ref]:
                logging.message('TRACE', 'Target ' + cls.file_log[file_ref]['target'] + ' is no longer part of the build, discontinuing tracking and deleting from build directory');
                if os.path.exists(cls.file_log[file_ref]['target']):
                    os.remove(cls.file_log[file_ref]['target'])
                cls.file_log.pop(file_ref)

        # remove build directories that are now empty
        empty_dirs = utilities.find_empty_subdirectories(config.system['paths']['profile_build_dir'])
        for dir in empty_dirs:
            logging.message('TRACE', 'Build directory ' + dir + ' no longer contains anything, deleting')
            os.rmdir(dir)
                
        # remove variables from the log that aren't referenced by a
        # file anymore
        variables_referenced = set()
        for file_entry in cls.file_log.values():
            if 'variables' in file_entry:
                for variable in file_entry['variables']:
                    variables_referenced.add(variable)
        for variable in cls.variable_change_flags:
            if variable not in variables_referenced:
                logging.message('TRACE', 'Variable ' + variable + ' is no longer part of the build, discontinuing tracking'); 
                cls.variable_log.pop(variable)

        # remove templates and resources from the log that aren't
        # referenced by a file anymore
        cls.remove_stale_references(cls.template_log, cls.template_change_flags, 'template')
        cls.remove_stale_references(cls.resource_log, cls.resource_change_flags, 'resource')

        # create the build log directory if it doesn't exist
        os.makedirs(os.path.join(config.system['paths']['config_root'], config.system['paths']['build_log_dir']), exist_ok = True)
        # write the file
        with open(cls.build_log_file, 'w') as build_log_fh:
            json.dump({'timestamp': cls.cur_build_time, 'files': cls.file_log, 'variables': cls.variable_log, 'templates': list(cls.template_log), 'resources': list(cls.resource_log)}, build_log_fh, default=str)
        
    #---------------------------------------------------------------------------
    
dependencies = DependencyManager



import collections.abc
import copy
import glob
import re
import os
from salal.core.logging import logging

class Utilities:

    #---------------------------------------------------------------------------
    @classmethod
    def deep_update (cls, target_dict, source_dict):
        # Updates <target_dict> with the values from <source_dict>,
        # overwriting existing values. If for a given key <key>,
        # <target_dict[key]> and <source_dict[key]> both exist and are
        # both some kind of Mapping, do a recursive update rather than
        # overwriting.
        for key, source_value in source_dict.items():
            target_value = target_dict.get(key)
            # note that this test will return false if the key isn't present
            # in the target dict, because target_value will be None 
            if isinstance(source_value, collections.abc.Mapping) and isinstance(target_value, collections.abc.Mapping):
                cls.deep_update(target_value, source_value)
            else:
                target_dict[key] = copy.deepcopy(source_value)
    
    #---------------------------------------------------------------------------

    @classmethod
    def find_subdirectories (cls, directory):
        # Returns a list of all subdirectories in the directory indicated by
        # <directory>. Unlike <find_files>, this method is non-recursive.
        result_list = []
        for entry in os.scandir(directory):
            if entry.is_dir():
                result_list.append(entry.name)
        return result_list

    #---------------------------------------------------------------------------

    @classmethod
    def find_files (cls, directory):
        # Recursively finds all files in the directory indicated by
        # <directory>. Returns a list of paths to these files relative
        # to <directory>.

        # For proper creation of a relative path, we need a consistent
        # convention regarding whether the directory has or does not
        # have a trailing separator. Here we've gone with 'has', and
        # add it if missing.
        if directory[-1] != os.sep:
            directory += os.sep
        # To use the directory name in an RE, we need to replace any backslashes
        # with double backslashes
        re_directory = directory.replace('\\', '\\\\')
        result_list = []
        # we need to look for files beginning with . specifically, as they
        # don't match * by default
        absolute_paths = glob.glob(directory + '**' + os.sep + '*', recursive = True) + glob.glob(directory + '**' + os.sep + '.*', recursive = True)
        for absolute_path in absolute_paths:
            if os.path.isfile(absolute_path):
                relative_path = re.match('^' + re_directory + '(.*)$', absolute_path).group(1)
                result_list.append(relative_path)
        return result_list

    #---------------------------------------------------------------------------

    @classmethod
    def find_files_by_extension (cls, directory, extension):
        # Recursively finds all files in the directory indicated by
        # <directory> that have the extension <extension>. Returns a list
        # of paths to these files relative to <directory>.

        # For proper creation of a relative path, we need a consistent
        # convention regarding whether the directory has or does not
        # have a trailing separator. Here we've gone with 'has', and
        # add it if missing.
        if directory[-1] != os.sep:
            directory += os.sep
        # To use the directory name in an RE, we need to replace any backslashes
        # with double backslashes
        re_directory = directory.replace('\\', '\\\\')
        result_list = []
        for absolute_path in glob.glob(directory + '**' + os.sep + '*.' + extension, recursive = True):
            relative_path = re.match('^' + re_directory + '(.*)$', absolute_path).group(1)
            result_list.append(relative_path)
        return result_list

    #---------------------------------------------------------------------------

    @classmethod
    def find_empty_subdirectories (cls, directory, result_list = None):
        # Recursively searches <directory> for any subdirectories that are
        # empty, and returns a list of those subdirectories.
        if result_list == None:
            result_list = []
        entry_count = 0
        for entry in os.scandir(directory):
            if entry.is_dir():
                cls.find_empty_subdirectories(os.path.join(directory, entry.name), result_list)
            entry_count += 1
        if entry_count == 0:
            result_list.append(directory)
        return result_list

    #---------------------------------------------------------------------------

    @classmethod
    def substitute_variables (cls, string, variables):
        # Jinja-style variable substitution. Any instances within
        # <string> where {{<identifier>}} occurs are replaced by
        # <variables[identifier]> if it exists, or an empty string
        # otherwise. Returns the resulting string. Any whitespace
        # around the identifier is ignored, so {{ <identifier> }} can
        # be used instead if it improves readability. If the variable
        # you need is in a dict inside of <variables>, you can access
        # it with dot notation, i.e., {{ foo.bar }} is equivalent to
        # variables['foo']['bar'].

        # split the string into literal text and variable references
        substrings = re.split('{{\s*(\S+)\s*}}', string)
        result = ''
        for index, substring in enumerate(substrings):
            # because of how re.split works, matching groups will be found
            # at odd-numbered indices in <substrings>
            if index % 2 == 1:
                # dots indicate a reference to a nested variable
                identifiers = substring.split('.')
                cur_dict = variables
                # if the identifier in a variable reference isn't in
                # <variables>, it gets replaced with the empty string
                match = ''
                for index, identifier in enumerate(identifiers):
                    if identifier not in cur_dict:
                        break
                    elif index < len(identifiers) - 1:
                        cur_dict = cur_dict[identifier]
                    else:
                        match = cur_dict[identifier]
                result += match 
            else:
                result += substring
        return result
            
    #---------------------------------------------------------------------------

utilities = Utilities



import sys
import os.path
import json
import argparse
from salal.core.logging import logging
from salal.core.utilities import utilities

# After initialization, the following attributes are available on
# the <config> object:
# - action: the action to be executed
# - profile: the build profile to use while executing it
# - system: a dict of system configuration variables
# - project: a dict of project configuration variables

class Config:

    #---------------------------------------------------------------------------

    @classmethod
    def initialize (cls):
        cls.set_salal_root()
        cls.parse_arguments()
        cls.load_site_configuration()
        cls.load_user_configuration()
        cls.load_build_profiles()
        cls.initialize_variables()
        cls.set_extension_directories()
        
    #---------------------------------------------------------------------------
 
    @classmethod
    def set_salal_root (cls):
        cls.system = {}
        cls.system['paths'] = {}
        cls.system['paths']['salal_root'] = os.path.normpath(os.path.dirname(sys.modules['__main__'].__file__))
    
    #---------------------------------------------------------------------------
   
    @classmethod
    def parse_arguments (cls):
        parser = argparse.ArgumentParser()
        parser.add_argument('action', action = 'store')
        parser.add_argument('profile', action = 'store', nargs = '?', default = 'default')
        parser.add_argument('--config-file', action = 'store', default = os.path.join(cls.system['paths']['salal_root'], 'site_config.json'))
        parser.add_argument('--logging-level', action = 'store', default = 'INFO')
        cls._arguments = parser.parse_args()
        # we shouldn't do any logging until this point has been reached,
        # otherwise it won't be impacted by the logging level
        logging.set_logging_level(cls._arguments.logging_level)
        cls.action = cls._arguments.action
        logging.message('DEBUG', 'Parsed command line arguments')

    #---------------------------------------------------------------------------

    @classmethod
    def load_site_configuration (cls):
        logging.message('DEBUG', 'Loading site configuration from ' + cls._arguments.config_file)
        with open(cls._arguments.config_file) as system_variables_fh:
            utilities.deep_update(cls.system, json.load(system_variables_fh))

    #---------------------------------------------------------------------------

    @classmethod
    def load_user_configuration (cls):
        config_file = os.path.expanduser(cls.system['paths']['user_config_file'])
        if os.path.isfile(config_file):
            logging.message('DEBUG', 'Loading user configuration from ' + config_file)
            with open(config_file) as system_variables_fh:
                utilities.deep_update(cls.system, json.load(system_variables_fh))

    #---------------------------------------------------------------------------

    @classmethod
    def load_build_profiles (cls):
        logging.message('DEBUG', 'Loading build profiles from ' + cls.system['paths']['profiles_file'])
        with open(cls.system['paths']['profiles_file']) as build_profiles_fh:
            cls._build_profiles = json.load(build_profiles_fh)
        
    #---------------------------------------------------------------------------

    @classmethod
    def initialize_variables (cls):
        logging.message('DEBUG', 'Using salal root directory of ' + cls.system['paths']['salal_root'])
        # convert the profile specifier to the correct profile name
        if cls._arguments.profile == 'default':
            cls.system['profile'] = None
            for build_profile in cls._build_profiles:
                if build_profile == 'common':
                    continue
                else:
                    cls.system['profile'] = build_profile
                    break
            if cls.system['profile'] == None:
                logging.message('ERROR', 'Default profile specified, but there are no profiles configured')
        elif cls._arguments.profile in cls._build_profiles:
            cls.system['profile'] = cls._arguments.profile
        else:
            logging.message('ERROR', 'Specified profile ' + cls._arguments.profile + ' does not exist')
        logging.message('INFO', 'Using profile ' + cls.system['profile'])
        cls.system['paths']['profile_build_dir'] = os.path.join(cls.system['paths']['build_root'], cls.system['profile'])
        
        logging.message('DEBUG', 'Initializing system and project variables')
        cls.project = dict()
        profile_vars = { 'system': cls.system, 'project': cls.project }
        for var_type in ['system', 'project']:
            if 'common' in cls._build_profiles and var_type in cls._build_profiles['common']:
                utilities.deep_update(profile_vars[var_type], cls._build_profiles['common'][var_type])
            if var_type in cls._build_profiles[cls.system['profile']]:
                utilities.deep_update(profile_vars[var_type], cls._build_profiles[cls.system['profile']][var_type])
        if 'theme_root' in config.system['paths']:
            logging.message('INFO', 'Using theme ' + config.system['paths']['theme_root'])
                
    #---------------------------------------------------------------------------

    @classmethod
    def set_extension_directories (cls):
        # Extensions can be located in three places: The base Salal
        # directory, the theme directory, or the <design> directory
        # for the project. In each case, any extensions need to be
        # placed in an <extensions> directory in that location. Here
        # we check for the existence of these <extensions> directories,
        # and set the system path <extension_dirs> to a list of those that
        # are found.
        extension_locations = [
            cls.system['paths']['salal_root'],
            cls.system['paths']['theme_root'] if 'theme_root' in config.system['paths'] else None,
            'design'
        ]
        config.system['paths']['extension_dirs'] = []
        for location in extension_locations:
            if location:
                extension_dir = os.path.join(location, cls.system['paths']['extensions_root'])
                if os.path.isdir(extension_dir):
                    config.system['paths']['extension_dirs'].append(extension_dir)
                    logging.message('DEBUG', 'Registered extensions directory ' + extension_dir)
        
    #---------------------------------------------------------------------------
    
config = Config



from .dependencies import dependencies

class VariableTracker(dict):

    #---------------------------------------------------------------------------
    
    def __init__(self, *args, success_callback, failure_callback, **kwargs):
        super().__init__(*args, **kwargs)
        self.success_callback = success_callback
        self.failure_callback = failure_callback
        
    #---------------------------------------------------------------------------

    def __getitem__(self, key):
        try:
            value = super().__getitem__(key)
            self.success_callback(key)
            return value
        except KeyError:
            self.failure_callback(key)

    #---------------------------------------------------------------------------



import os
from salal.core.logging import logging
from salal.core.config import config
from salal.core.handlers import handlers
from salal.core.dependencies import dependencies

class FileProcessing:

    #---------------------------------------------------------------------------

    @classmethod
    def initialize (cls):
        logging.message('DEBUG', 'Loading file processing handlers')
        cls.handlers = handlers.load_handlers(config.system['paths']['file_processing_handlers_dir'])
        
    #---------------------------------------------------------------------------

    @classmethod
    def process (cls, source_dir, target_dir, file_relative_path):
        file_stem, ext = os.path.splitext(file_relative_path)
        if file_stem.startswith('.'):
            ext = file_stem
            file_stem = ''
        # strip off the initial '.' in the extension so we just have letters
        ext = ext[1:]
        if ext in cls.handlers:
            tag = ext
        elif 'default' in cls.handlers:
            tag = 'default'
        else:
            logging.message('WARN', 'Handling for file type ' + ext + ' is not configured, skipping.')
            return
        target_ext = cls.handlers[tag].get_target_extension(ext)
        source_file = os.path.join(source_dir, file_relative_path)
        target_file = os.path.join(target_dir, file_stem + '.' + target_ext)
        if dependencies.needs_build(target_file, source_file):
            # create the target directory if it doesn't exist
            os.makedirs(os.path.join(target_dir, os.path.dirname(file_relative_path)), exist_ok = True)
            logging.message('INFO', source_file)
            dependencies.start_build_tracking(target_file, source_file)
            cls.handlers[tag].process(ext, source_dir, target_dir, file_stem)
            dependencies.stop_build_tracking()

    #---------------------------------------------------------------------------

file_processing = FileProcessing
